import HellbornActorSheet from "./base-actor-sheet.mjs"

export default class HellbornEnemySheet extends HellbornActorSheet {
  /** @override */
  static DEFAULT_OPTIONS = {
    classes: ["enemy"],
    position: {
      width: 860,
      height: 620,
    },
    window: {
      contentClasses: ["enemy-content"],
    },
    actions: {
      createTrait: HellbornEnemySheet.#onCreateTrait,
      createEquipment: HellbornEnemySheet.#onCreateEquipment,
      createWeapon: HellbornEnemySheet.#onCreateWeapon,
      createMalefica: HellbornEnemySheet.#onCreateMalefica,
    }

  }

  /** @override */
  static PARTS = {
    main: {
      template: "systems/fvtt-hellborn/templates/enemy-main.hbs",
    },
    tabs: {
      template: "templates/generic/tab-navigation.hbs",
    },
    traits: {
      template: "systems/fvtt-hellborn/templates/enemy-trait.hbs",
    },
    biography: {
      template: "systems/fvtt-hellborn/templates/enemy-biography.hbs",
    },
  }

  /** @override */
  tabGroups = {
    sheet: "traits",
  }


  /**
   * Prepare an array of form header tabs.
   * @returns {Record<string, Partial<ApplicationTab>>}
   */
  #getTabs() {
    const tabs = {
      traits: { id: "traits", group: "sheet", icon: "fa-solid fa-shapes", label: "HELLBORN.Label.traits" },
      biography: { id: "biography", group: "sheet", icon: "fa-solid fa-book", label: "HELLBORN.Label.biography" },
    }
    for (const v of Object.values(tabs)) {
      v.active = this.tabGroups[v.group] === v.id
      v.cssClass = v.active ? "active" : ""
    }
    return tabs
  }

  /** @override */
  async _prepareContext() {
    const context = await super._prepareContext()
    context.tabs = this.#getTabs()

    return context
  }

  /** @override */
  async _preparePartContext(partId, context) {
    const doc = this.document
    switch (partId) {
      case "main":
        break
      case "traits":
        context.tab = context.tabs.traits
        context.traits = doc.itemTypes["species-trait"]
        context.traits.sort((a, b) => a.name.localeCompare(b.name))
        context.weapons = doc.itemTypes.weapon
        context.weapons.sort((a, b) => a.name.localeCompare(b.name))
        context.maleficas = doc.itemTypes.malefica
        context.maleficas.sort((a, b) => a.name.localeCompare(b.name))
        context.equipments = doc.itemTypes.equipment
        context.equipments.sort((a, b) => a.name.localeCompare(b.name))
        break
      case "biography":
        context.tab = context.tabs.biography
        context.enrichedDescription = await foundry.applications.ux.TextEditor.implementation.enrichHTML(doc.system.description, { async: true })
        context.enrichedNotes = await foundry.applications.ux.TextEditor.implementation.enrichHTML(doc.system.notes, { async: true })
        break
    }
    return context
  }

  /**
   * Creates a new attack item directly from the sheet and embeds it into the document.
   * @param {Event} event             The initiating click event.
   * @param {HTMLElement} target      The current target of the event listener.
   */
  static #onCreateTrait(event, target) {
    this.document.createEmbeddedDocuments("Item", [{ name: game.i18n.localize("HELLBORN.Label.newTrait"), type: "species-trait" }])
  }

  static #onCreateEquipment(event, target) {
    this.document.createEmbeddedDocuments("Item", [{ name: game.i18n.localize("HELLBORN.Label.newEquipment"), type: "equipment" }])
  }

  static #onCreateMalefica(event, target) {
    this.document.createEmbeddedDocuments("Item", [{ name: game.i18n.localize("HELLBORN.Label.newMalefica"), type: "malefica" }])
  }

  static #onCreateWeapon(event, target) {
    this.document.createEmbeddedDocuments("Item", [{ name: game.i18n.localize("HELLBORN.Label.newWeapon"), type: "weapon" }])
  }


  /**
   * Handles the roll action triggered by user interaction.
   *
   * @param {PointerEvent} event The event object representing the user interaction.
   * @param {HTMLElement} target The target element that triggered the roll.
   *
   * @returns {Promise<void>} A promise that resolves when the roll action is complete.
   *
   * @throws {Error} Throws an error if the roll type is not recognized.
   *
   * @description This method checks the current mode (edit or not) and determines the type of roll
   * (save, resource, or damage) based on the target element's data attributes. It retrieves the
   * corresponding value from the document's system and performs the roll.
   */
  async _onRoll(event, target) {
    const rollType = $(event.currentTarget).data("roll-type")
    let item
    switch (rollType) {
      case "stat":
        {
          let statId = $(event.currentTarget).data("stat-id");
          item = this.actor.system.stats[statId];
          await this.document.system.roll(rollType, item)
        }
        break
      case "weapon":
      case "damage":
        {
          let li = $(event.currentTarget).parents(".item");
          item = this.actor.items.get(li.data("item-id"));
          await this.document.system.roll(rollType, item)
        }
        break
      default:
        throw new Error(`Unknown roll type ${rollType}`)
    }
  }

  async _onDrop(event) {
    if (!this.isEditable || !this.isEditMode) return
    const data = TextEditor.getDragEventData(event)

    // Handle different data types
    switch (data.type) {
      case "Item":
        const item = await fromUuid(data.uuid)
        return super._onDropItem(item)
    }
  }
}
